/* * (C) Copyright IBM Corporation 2015, 2016. */
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA


 Author: IBM Corporation
 This module is part of the IBM (R) Rational (R) ClearCase (R)
 Multi-version file system (MVFS).
 For support, please visit http://www.ibm.com/software/support

*/

/* Description:
**
** Three mnode classes come here for their lookup behavior.  The MFS_NTVWCLAS
** with the auto_view flag set, MFS_VOBTAGCLAS (only available in automatic
** views), and MFS_VOBRTCLAS mounted in an automatic view for the special case
** of looking up "..".  Only automatic MFS_NTVWCLAS and MFS_VOBTAGCLAS come here
** for readdir and getattr behavior.  For a description of the MFS_NTVWCLAS
** without the auto_view flag set, see mvfs_ntvwops.c.  For other lookups
** (besides "..") in a MFS_VOBRTCLAS (in an automatic view or not), see
** mvfs_vnodeops.c.
**
** Automatic views only support view-extended naming and per-view VOB mounting
** (through the mount ioctl).  Unlike in a dynamic view for the MFS_NTVWCLAS, a
** vobtag for a VOB mount in an automatic view can have multiple components
** (like a vobtag on a Unix system).  For instance, a vob tag could be
** "/var/tmp/<vob-name>".  In a Unix dynamic view this would be handled by
** putting a real filesystem mountpoint at "/var/tmp/<vob-name>".  However, we
** don't have privileged access for automatic views, so we can't create a real
** filesystem mountpoint.  Instead, we use the mount ioctl to mount that tag
** under the view such that each component of the vobtag shows up as a "virtual
** directory" under the view.  This gives VOB mount pathnames like this in an
** automatic view:
**
**   pathname                                kind of MVFS object
**   -----------------------------------------------------------
**   /rview (or r:\)                         MFS_VIEWDIRCLAS
**   /rview/<view-tag>                       MFS_NTVWCLAS
**   /rview/<view-tag>/<vob-name>            MFS_VOBRTCLAS (single-component)
**   /rview/<view-tag>/var                   MFS_VOBTAGCLAS
**   /rview/<view-tag>/var/tmp               MFS_VOBTAGCLAS
**   /rview/<view-tag>/var/tmp/<vob-name>    MFS_VOBRTCLAS (multi-component)
**
** Most operations on these VOB tag virtual directories return EINVAL or EROFS,
** since the directory the tag represents can not be written to.
**
** The following read-only operations are supported and implemented in this
** module:
**
**   getattr()  Gets (faked) stats for the vobtag component.
**   lookup()   Attempts to lookup a VOBRT or VOBTAG object for this view.
**   readdir()  Returns a list of VOBRT and/or VOBTAG components in the readdir
**              directory entry format.
*/
#include "mvfs_systm.h"
#include "mvfs.h"

/* The maximum number of components we'll allow in a vobtag. */
enum {
    MVFS_MAX_VOBTAG_COMP = 64
};

/****************************************************************************
 * mvfs_auvw_lookup_vpidx
 * Lookup the entry index for vp.
 * IN   vp              Vnode to lookup
 * OUT  vpidxp          Index of entry with that vp
 * RESULT:              Unix errno
 *
 * Well, this is annoying...since our array is sorted on parent_vp and then nm,
 * we have to search linearly for the vp.  Caller manages MVFS_AUVW_VOBTAG_LOCK.
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_lookup_vpidx(
    VNODE_T *vp,
    int *vpidxp
)
{
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(vp)));
    int error;
    int idx;

    for (idx = 0; idx < mvtdp->vobtag_dir_count; idx++) {
        /* Make sure it's a real entry and not one for a "..". */
        if ((mvtdp->vobtag_dir_ents[idx].vobtag_vp == vp) &&
            (STRCMP(mvtdp->vobtag_dir_ents[idx].vobtag_nm, "..") != 0))
        {
            *vpidxp = idx;
            error = 0;
            goto done;
        }
    }
    error = ENOENT;
  done:
    return error;
}

/****************************************************************************
 * mvfs_auvw_lookup_subr
 * Lookup nm in dir, return its vnode, or insert index (if ENOENT).
 * IN   dkey            Dir vp and name we're looking up
 * IO   pnp             Ptr to a pathname struct (may be NULL)
 * OUT  vpp             Resulting vp or NULL if the lookup fails (may be NULL)
 * OUT  idxp            Insertion idx if ENOENT, else unset (may be NULL)
 * RESULT:              Unix errno
 *
 *  We handle case insensitive names and expect the caller to manage
 *  MVFS_AUVW_VOBTAG_LOCK.
 *
 * Lookup the name in the struct mvfs_vobtag_dir for the view associated with
 * dvp.  The dvp could be the view itself, a vobtag, or a vobrt component (if
 * we're looking up "..").  The MFS_VIEW(dvp) works on view, vobtag, or vobrt
 * vnodes to return the associated view.  Do a binary search with inclusive
 * bounds.  Also, use the trick that (low + high) / 2 might overflow due to the
 * addition (not really in our case, but deal with it anyway), but it is
 * equivalent to low + ((high - low) / 2), which can't overflow.  Note, the
 * array is sorted on the primary key, which is the parent vp, and then
 * secondarily sorted on the name, so the comparisons need to take that into
 * account.  Also note, we never put in an entry with a NULL name (which would
 * cause a panic in STRCMP) so we don't have to check for that.  Here's a little
 * picture of a representative search (the first row is the sorted parent_vp's
 * and the second row is the names, sorted per parent_vp):
 *
 *                                                              high-water
 *                2nd     3rd 4th  1st                               |
 *                 |       |   |    |                                |
 *                 v       v   v    v                                v
 * +-----------+-----+-------------------------+-----+---------------+
 * |           |     |                         |     |               |
 * +-----------+-----+-------------------------+-----+---------------+
 * | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | | |
 * +-----------+-----+-------------------------+-----+---------------+
*
 * 1st: parent_vp =, name <
 * 2nd: parent_vp <, name not checked
 * 3rd: parent_vp =, name >
 * 4th: parent_vp =, name =
 *
 * Our database is kept as a compact sorted array.  That way, we can use binary
 * searches for both lookup and readdir.  This could be somewhat slower than a
 * hashtable for the lookup case (if we had a really good hash and a big enough
 * table), but is considerably faster for the readdir case (which otherwise
 * would have to do a linear search through the database looking for matching
 * parent_vp entries).  Probably lots more algorithm analysis could be done, but
 * this is fairly simple and should do well enough.  The array is maintained
 * "compactly", so inserting or removing entries requires opening or closing a
 * space in the array (by moving all higher entries "left" or "right".  VOB
 * mounting and unmounting is not very common so the expense of moving memory
 * shouldn't be noticable.
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_lookup_subr(
    mvfs_vobtag_dir_ent_key_t dkey,
    struct pathname *pnp,
    VNODE_T **vpp,
    int *idxp
)
{
    VNODE_T *dvp = dkey.vtk_parent_vp;
    char *nm = dkey.vtk_nm;
    tbs_boolean_t case_insensitive = MVFS_PN_CI_LOOKUP(pnp);
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp)));
    int error;

    VNODE_T *vp = NULL;         /* We haven't found a match yet. */
    int idx;                    /* Match idx or insert idx. */
    int low = 0;
    int high = (int)(mvtdp->vobtag_dir_count) - 1;
    int mid;
    int cmp;

    /* Do the binary search.  Note, the high index could start negative (if
    ** vobtag_dir_count is 0 for an empty array, which is why we need the (int)
    ** cast above) or become negative (if mid is 0 and we don't match in the
    ** search).  That's OK because we will end the search at that point and
    ** never use it as an array index.
    */
    while (low <= high) {
        mid = low + ((high - low) / 2);

        /* Check the primary key. */
        if (mvtdp->vobtag_dir_ents[mid].vobtag_parent_vp == dvp) {
            /* Primary key matches, check the secondary key. */
            cmp = PN_STRCMP(case_insensitive,
                            mvtdp->vobtag_dir_ents[mid].vobtag_nm,
                            nm);

            if (cmp == 0) {
                /* Both keys match, so we found it and we're done. */
                vp = mvtdp->vobtag_dir_ents[mid].vobtag_vp;
                break;
            } else if (cmp < 0) {
                low = mid + 1;
            } else {
                high = mid - 1;
            }
        } else if (mvtdp->vobtag_dir_ents[mid].vobtag_parent_vp < dvp) {
            low = mid + 1;
        } else {
            high = mid - 1;
        }
    }
    /* Check the results of our search. */
    if (vp == NULL) {
        idx = low;              /* Index of insertion point. */
        error = ENOENT;
    } else {
        /* We found a vp to match the name, we hold it below if we're returning
        ** it to the caller.  Also, set the index that we found (the caller may
        ** want it, probably to manage the refcnt).  If the caller wanted a
        ** case-insensitive lookup we still return a case-correct component.
        */
        if (case_insensitive && (pnp != NULL)) {
            if ((error = PN_SET_CASE_CORRECT_COMP(pnp, nm)) != 0) {
                /* We got an error, but being unable to set the correct
                ** case doesn't seem that bad, so just return what we found and
                ** hope the caller isn't confused.
                */
                mvfs_log(MFS_LOG_ERR,
                         "mvfs_auvw_lookup_comp_subr: error=%d from"
                         " PN_SET_CASE_CORRECT_COMP(pnp=%p, nm=%s)"
                         " continuing...\n",
                         error, pnp, nm);
            }
        }
        idx = mid;              /* Index of matching entry. */
        error = 0;
    }
    /* If caller wants it, return the held vp we found or NULL. */
    if (vpp != NULL) {
        if (vp != NULL) {
            VN_HOLD(vp);
        }
        *vpp = vp;
    }
    /* If caller wants it, return the index of insertion point or match. */
    if (idxp != NULL) {
        *idxp = idx;
    }
    return error;
}

/****************************************************************************
 * mvfs_auvw_lookup_comp
 * Lookup nm in dir, return its vnode.  Caller manages MVFS_AUVW_VOBTAG_LOCK.
 * IN   dkey            Dir vp and name we're looking up
 * IO   pnp             Ptr to a pathname struct (may be NULL)
 * OUT  vpp             Resulting vp or NULL if the lookup fails (may be NULL)
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_lookup_comp(
    mvfs_vobtag_dir_ent_key_t dkey,
    struct pathname *pnp,
    VNODE_T **vpp
)
{
    /* We don't care about the index since this is a lookup for vp. */
    return mvfs_auvw_lookup_subr(dkey, pnp, vpp, NULL);
}

/****************************************************************************
 * mvfs_auvw_lookup_comp_idx
 * Lookup nm in dir, return its vnode, or insert index (if ENOENT).  Caller
 * manages MVFS_AUVW_VOBTAG_LOCK.
 * IN   dkey            Dir vp and name we're looking up
 * OUT  idxp            Index for insertion if ENOENT, else index found
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_lookup_comp_idx(
    mvfs_vobtag_dir_ent_key_t dkey,
    int *idxp
)
{
    /* This lookup is always case-sensitive (hence the NULL for pnp) and we
    ** don't need the vnode (hence the other NULL), just the index.
    */
    return mvfs_auvw_lookup_subr(dkey, NULL, NULL, idxp);
}

/****************************************************************************
 * mvfs_auvw_insert_entry
 * Insert an entry into the database.
 * IN   mvtdp           The database pointer
 * IN   vpidx           Index of entry to add
 * IN   dkey            Dir vp and name we're adding
 * IN   vp              Vnode for the name we're adding
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_insert_entry(
    mvfs_vobtag_dir_t *mvtdp,
    int vpidx,
    mvfs_vobtag_dir_ent_key_t dkey,
    VNODE_T *vp
)
{
    mvfs_vobtag_dir_ent_t *mvtdep;
    mvfs_vobtag_dir_ent_t saved_ent; /* For cleanup. */
    int ents_to_move;
    int error = 0;

    if (mvtdp->vobtag_dir_count >= MVFS_VOBTAG_DIR_ENTS) {
        error = ENOMEM;         /* One more will be too many. */
        goto done;
    }
    /* Open up the entry at vpidx by moving the rest of the array to the right
    ** (memmove allows overlapping memory areas).  We're going to move (for a
    ** zero-based array) "vobtag_dir_count - vpidx" entries, which might be none
    ** if there aren't any entries in the array yet or the insertion point is
    ** the right end of the array.  We don't care about the value memmove
    ** returns (it's just the dst ptr).
    */
    if ((ents_to_move = mvtdp->vobtag_dir_count - vpidx) > 0) {
        (void)MDKI_MEMMOVE((void *)(&((mvtdp->vobtag_dir_ents)[vpidx + 1])),
                           (void *)(&((mvtdp->vobtag_dir_ents)[vpidx])),
                           ents_to_move * sizeof(mvfs_vobtag_dir_ent_t));
    }
    /* The array is now one entry bigger. */
    mvtdp->vobtag_dir_count += 1;

    /* Point to the entry to fill in.  We're putting this stuff in our database,
    ** so copy the name and hold vp's.
    */
    mvtdep = &((mvtdp->vobtag_dir_ents)[vpidx]);
    saved_ent = *mvtdep;        /* In case we need to cleanup. */

    if ((mvtdep->vobtag_key.vtk_nm =
         (char *)KMEM_ALLOC(STRLEN(dkey.vtk_nm) + 1, KM_SLEEP)) == NULL)
    {
        /* Oops, put the array back the way it was. */
        if (ents_to_move > 0) {
            (void)MDKI_MEMMOVE((void *)(&((mvtdp->vobtag_dir_ents)[vpidx])),
                               (void *)(&((mvtdp->vobtag_dir_ents)[vpidx + 1])),
                               ents_to_move * sizeof(mvfs_vobtag_dir_ent_t));
        }
        *mvtdep = saved_ent;
        mvtdp->vobtag_dir_count -= 1;
        error = ENOMEM;
        goto done;
    }
    STRCPY(mvtdep->vobtag_key.vtk_nm, dkey.vtk_nm);
    mvtdep->vobtag_key.vtk_parent_vp = dkey.vtk_parent_vp;
    VN_HOLD(mvtdep->vobtag_key.vtk_parent_vp);
    mvtdep->vobtag_vp = vp;
    VN_HOLD(mvtdep->vobtag_vp);
    mvtdep->vobtag_refcnt = 1;  /* One refcnt for itself. */

  done:
    return error;
}

/****************************************************************************
 * mvfs_auvw_insert_comp
 * Insert an entry for a component into the database, including its "..".
 * IN   pdkey           Parent vp and dir name of dir we're adding dkey to
 * IN   dkey            Dir vp and name we're adding
 * IN   vp              Vnode for the name we're adding
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_insert_comp(
    mvfs_vobtag_dir_ent_key_t pdkey,
    mvfs_vobtag_dir_ent_key_t dkey,
    VNODE_T *vp
)
{
    VNODE_T *dvp = dkey.vtk_parent_vp;
    char *nmp = dkey.vtk_nm;
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp)));
    mvfs_vobtag_dir_ent_t *mvtdep;
    int ents_to_move;
    int vpidx;
    int dvpidx;
    int error;

    /* Our caller manages this lock. */
    ASSERT(MVFS_AUVW_VOBTAG_ISLOCKED(mvtdp));

    /* The parent dvp must be a vobtag, an NT auview, or a vobrt. */
    ASSERT(MFS_ISVOBTAG(VTOM(dvp)) ||
           MFS_ISAUVIEW(VTOM(dvp)) ||
           MFS_ISVOBRT(VTOM(dvp)));

    /* We don't allow NULL ptrs for names (STRCPY or STRCMP would panic). */
    if (nmp == NULL) {
        error = EINVAL;
        goto done;
    }
    /* Find the index of the insertion point for this name (which better not be
    ** there since we shouldn't be here in insert if it is) and insert it.
    */
    if ((error = mvfs_auvw_lookup_comp_idx(dkey, &vpidx)) != ENOENT) {
        /* error could be 0, which is also an error ;). */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_insert_comp: err=%d (should be ENOENT) looking up"
                 " idx of nm=%s in dnm=%s (dvp=%p)\n",
                 nmp, (pdkey.vtk_nm != NULL) ? pdkey.vtk_nm : "???", dvp);
        goto done;
    }
    if ((error = mvfs_auvw_insert_entry(mvtdp, vpidx, dkey, vp)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_insert_comp: Inserting %s (dvp=%p) at idx=%d"
                 " err=%d\n",
                 nmp, dvp, vpidx, error);
        goto done;
    }
    /* Insert ".." into the new dir.  The dvp we started with will be the
    ** "answer" to ".." (could be a MFS_NTVWCLAS or a MFS_VOBTAGCLAS vnode).
    ** Since the ".." entry is "tied" to this entry (i.e., it is created here
    ** when the entry is created and it is removed when the entry is removed,
    ** and it only ever has a refcnt of 1 (never has any subdirs)), it would be
    ** nice if we could just keep the info in this entry and not create another
    ** entry for "..".  However, we have to do lookups of ".." given only a dvp,
    ** not a name, so we need an entry for ".." to be able to find it...Oh,
    ** well.
    **
    ** Note, we don't put in a "." since it is trivial for mvfs_auvw_lookup_comp
    ** (just the current dir vp) and we special case it in mvfs_auvw_readdir
    ** since we don't have to return anything from the database there.
    **
    ** Note, the dkey arg is passed by value, so it's OK to overwrite it here.
    */
    dkey.vtk_parent_vp = vp;
    dkey.vtk_nm = "..";

    /* Increment our entry refcnt now because the entry could move when we
    ** insert the ".." entry, and this saves looking it up again.  We can put it
    ** back on an error because then it won't have moved.
    */
    dvpidx = vpidx;             /* Save the index in case of cleanup. */
    mvtdp->vobtag_dir_ents[dvpidx].vobtag_refcnt += 1;

    if ((error = mvfs_auvw_lookup_comp_idx(dkey, &vpidx)) != ENOENT) {
        /* error could be 0, which is also an error ;). */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_insert_comp: err=%d (should be ENOENT) looking up"
                 " idx of .. in nm=%s (vp=%p)\n",
                 error, nmp, vp);
        mvtdp->vobtag_dir_ents[dvpidx].vobtag_refcnt -= 1;
        goto done;
    }
    if ((error = mvfs_auvw_insert_entry(mvtdp, vpidx, dkey, dvp)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_insert_comp: Inserting .. in %s (dvp=%p) at idx=%d"
                 " err=%d\n",
                 nmp, dvp, vpidx, error);
        mvtdp->vobtag_dir_ents[dvpidx].vobtag_refcnt -= 1;
        goto done;
    }
    /* We just added a child of dvp (the parent of the entry we're adding), so
    ** increment its refcnt.  We have to lookup the index of the dvp to do that
    ** (which is why we need its parent vp and name in the pdkey).  Note, if the
    ** dvp is the view then we don't have an entry for it, so there's no sense
    ** looking (the lookup would return ENOENT in that case, which we could
    ** check for, but the test is faster).
    */
    if (MFS_ISAUVIEW(VTOM(dvp))) {
        goto done;
    }
    /* Since it's not a view, it must be a MFS_VOBTAGCLAS or a MFS_VOBRTCLAS and
    ** the pdkey should be filled in.
    */
    ASSERT((pdkey.vtk_parent_vp != NULL) && (pdkey.vtk_nm != NULL));

    if ((error = mvfs_auvw_lookup_comp_idx(pdkey, &vpidx)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_insert_comp: Lookup idx of dvp=%p for nm=%s in"
                 " pvp=%p dnm=%s err=%d\n",
                 dvp, nmp, pdkey.vtk_parent_vp, pdkey.vtk_nm, error);
        goto done;
    }
    mvtdp->vobtag_dir_ents[vpidx].vobtag_refcnt += 1;

  done:
    return error;
}

/****************************************************************************
 * mvfs_auvw_remove_entry
 * Remove an entry in the database.  Always succeeds.
 * IN   mvtdp           The database pointer
 * IN   vpidx           Index of entry to remove
 * IN   cdp             Call data (includes user cred)
 */
STATIC void MVFS_NOINLINE
mvfs_auvw_remove_entry(
    mvfs_vobtag_dir_t *mvtdp,
    int vpidx,
    CALL_DATA_T *cdp
)
{
    mvfs_vobtag_dir_ent_t *mvtdep = &((mvtdp->vobtag_dir_ents)[vpidx]);
    int ents_to_move;

    /* Free up stuff in the entry. */
    KMEM_FREE(mvtdep->vobtag_nm, STRLEN(mvtdep->vobtag_nm) + 1);
    ATRIA_VN_RELE(mvtdep->vobtag_parent_vp, cdp);
    ATRIA_VN_RELE(mvtdep->vobtag_vp, cdp);

    /* Now "remove" the entry by "shifting left" (which overwrites it).  If this
    ** is the "right-most" entry (which could be the only entry), no movement is
    ** necessary.  We're removing an entry, so we're guaranteed that count is at
    ** least 1, so the - 1 here and below is OK (never goes "negative").
    */
    if ((ents_to_move = (mvtdp->vobtag_dir_count - 1) - vpidx) > 0) {
        (void)MDKI_MEMMOVE((void *)(&((mvtdp->vobtag_dir_ents)[vpidx])),
                           (void *)(&((mvtdp->vobtag_dir_ents)[vpidx + 1])),
                           ents_to_move * sizeof(mvfs_vobtag_dir_ent_t));
    }
    /* Initialize the leftover (right-most, maybe only) entry.  This isn't
    ** strictly necessary since we should never look beyond the current
    ** vobtag_dir_count (which we decrement below), but do it anyway for
    ** "cleanliness".
    */
    mvtdep = &((mvtdp->vobtag_dir_ents)[mvtdp->vobtag_dir_count - 1]);
    mvtdep->vobtag_parent_vp = NULL;
    mvtdep->vobtag_nm = NULL;
    mvtdep->vobtag_vp = NULL;
    mvtdep->vobtag_refcnt = 0;

    mvtdp->vobtag_dir_count -= 1; /* May end at 0, an empty array. */
}

/****************************************************************************
 * mvfs_auvw_remove_comp
 * Remove the entry for a component in the database, including its "..".
 * IN   vp              Vnode pointer of the entry to remove
 * IN   vpidx           Index of entry to remove
 * IN   cdp             Call data (includes user cred)
 * OUT  rmvdp           TRUE - entry removed, FALSE - entry refcnt'd
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_remove_comp(
    VNODE_T *vp,
    int vpidx,
    CALL_DATA_T *cdp,
    tbs_boolean_t *rmvdp
)
{
    int error = 0;              /* Currently, never fails. */
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(vp)));
    mvfs_vobtag_dir_ent_key_t dkey;

    /* Our caller manages this lock, and do a sanity check on the inputs. */
    ASSERT(MVFS_AUVW_VOBTAG_ISLOCKED(mvtdp) &&
           (vpidx >= 0) &&
           (vpidx < mvtdp->vobtag_dir_count) &&
           (mvtdp->vobtag_dir_ents[vpidx].vobtag_vp == vp));

    /* If the entry has more refs than for itself and "..", we're done. */
    if (--(mvtdp->vobtag_dir_ents[vpidx].vobtag_refcnt) > 2) {
        *rmvdp = FALSE;
        goto done;
    }
    /* Otherwise, remove the entry.  Note, this will rele the vp for this entry,
    ** but it still has a hold because it is the parent vp for the ".." entry
    ** (and we'll need it to do that lookup).
    */
    mvfs_auvw_remove_entry(mvtdp, vpidx, cdp);
    *rmvdp = TRUE;

    /* Remove the ".." entry for the entry we just removed. */
    dkey.vtk_parent_vp = vp;
    dkey.vtk_nm = "..";

    if ((error = mvfs_auvw_lookup_comp_idx(dkey, &vpidx)) != 0) {
        /* The ".." better be there. */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_remove_comp: lookup .. vp=%p err=%d\n",
                 vp, error);
        goto done;
    }
    ASSERT(mvtdp->vobtag_dir_ents[vpidx].vobtag_refcnt == 1);
    mvfs_auvw_remove_entry(mvtdp, vpidx, cdp);

  done:
    return error;
}

/****************************************************************************
 * mvfs_auvw_mount_comp
 * Find or create "directory" for cnmp, including ".." under it.  The caller
 * manages the MVFS_AUVW_VOGTAG_LOCK.
 * IN   vfsp            VFS for the VOB root we're mounting
 * IN   pdkey           Parent vp and dir name of dir we're adding dkey to
 * IN   dkey            Dir vp and name we're adding
 * IN   lc              Last component, i.e., VOBRT (TRUE) or not (FALSE)
 * IN   cdp             Call data (includes user cred)
 * OUT  vpp             Held vnode created for the component
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_mount_comp(
    VFS_T *vfsp,
    mvfs_vobtag_dir_ent_key_t pdkey,
    mvfs_vobtag_dir_ent_key_t dkey,
    tbs_boolean_t lc,
    CALL_DATA_T *cdp,
    VNODE_T **vpp
)
{
    VNODE_T *dvp = dkey.vtk_parent_vp;
    char *nmp = dkey.vtk_nm;
    VNODE_T *vp;                /* The vp we will return. */
    int error;                  /* Error we will return. */

    /* Should only be called correctly (not a complete check of args). */
    ASSERT(dvp != NULL && nmp != NULL && vpp != NULL);

    if ((error = mvfs_auvw_lookup_comp(dkey, NULL, &vp)) == 0) {
        /* We found it, just return (vp is held by the lookup). */
        goto done;
    }
    if (error != ENOENT) {
        /* Some bad error, just return (vp was set to NULL on error). */
        goto done;
    }
    /* The lookup didn't find it, so create a new one.  We're still under the
    ** lock (our caller holds it) so the lookup failure is still valid.  We're
    ** doing this mount of components after we've already done the mvfs_vmount
    ** (called from mvfs_register_mount) and that has done a mfs_makevobrtnode
    ** for the VOB root to create a root vnode (and MFS_VOBRTCLAS mnode along
    ** with it).  If this is the last component of the multi-component vobtag
    ** (which might only be a single component), then we need an entry in our
    ** database for it so we can handle a ".." lookup, which will either be the
    ** "parent" component (MFS_VOBTAGCLAS) or the (automatic) view
    ** (MFS_NTVWCLAS).  Also, since we didn't find this in the lookup, we need a
    ** new mnode so we pass the fidp as NULL to mfs_makespecnode.
    */
    if (lc) {
        /* For a vobroot, use the VFS rootvp to add to our database. */
        ASSERT(MFS_ISVOBRT(VTOM(VFS_TO_MMI(vfsp)->mmi_rootvp)));
        vp = VFS_TO_MMI(vfsp)->mmi_rootvp;
        VN_HOLD(vp);
    } else {
        /* Make a vobtag class vnode to add to our database.  Since these vnodes
        ** aren't really tied to a vob VFS (they last as long as vobtag entries
        ** might include them), they should be mmi_refcnt'd (which this vnode
        ** creation does) by the view VFS.  For instance, two vobs with tags
        ** /vobs/foo and /vobs/bar will only have one entry for "vobs" and it
        ** needs to stay there until both vobs are unmounted, so the mmi_refcnt
        ** for the VOB VFS's can't include it (or they won't unmount because
        ** their mmi_refcnt won't be 1).
        **
        ** The mn_hdr will be set up by mvfs_mninit_new_mnode called by
        ** mfs_makespecnode.  In particular, the mn_hdr.viewvp is set to the
        ** view (and held) we passed in, and the mn_hdr.fid is "created" (since
        ** we passed in NULL).
        */
        if ((error = mfs_makespecnode(MFS_VOBTAGCLAS,
                                      MFS_VIEW(dvp),
                                      NULL, /* fidp */
                                      MFS_VIEW(dvp)->v_vfsp,
                                      &vp,
                                      cdp)) != 0)
        {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_mount_comp: makespecnode for nm=%s (vp=%p)"
                     " in dnm=%s (vp=%p) err=%d\n",
                     dkey.vtk_nm, vp, pdkey.vtk_nm, pdkey.vtk_parent_vp, error);
            vp = NULL;          /* Just to be sure. */
            goto done;
        }
    }
    /* Create an entry for the vp we just got/created (and its ".."). */
    if ((error = mvfs_auvw_insert_comp(pdkey, dkey, vp)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_mount_comp: Insert of nm=%s (vp=%p) in dnm=%s"
                 " (vp=%p) err=%d\n",
                 dkey.vtk_nm, vp, pdkey.vtk_nm, pdkey.vtk_parent_vp, error);
        /* The vp will be rele'd below because of the error. */
        goto done;
    }
  done:
    /* On error, we need to rele the vp if we have one. */
    if (error != 0) {
        if (vp != NULL) {
            ATRIA_VN_RELE(vp, cdp);
            vp = NULL;
        }
    }
    /* The vp was held above (by lookup or by us), or it is NULL. */
    *vpp = vp;
    return error;
}

/****************************************************************************
 * mvfs_auvw_unmount_comp
 * Remove the entry for cnmp, including ".." under it and parent(s) over it.
 * IN   dkey            Dir vp and name we're removing
 * IN   lc              Last component, i.e., VOBRT (TRUE) or not (FALSE)
 * IN   cdp             Call data (includes user cred)
 * OUT  vpp             Held pointer to vnode for cnmp (for next dvp)
 * RESULT:              Unix errno
 *
 * The caller manages the MVFS_AUVW_VOGTAG_LOCK.
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_unmount_comp(
    mvfs_vobtag_dir_ent_key_t dkey,
    tbs_boolean_t lc,
    CALL_DATA_T *cdp,
    VNODE_T **vpp
)
{
    VNODE_T *dvp = dkey.vtk_parent_vp;
    char *nmp = dkey.vtk_nm;
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp)));
    VNODE_T *vp = NULL;         /* The vp we will return. */
    int vpidx;
    VNODE_T *lvp;               /* Local vp for "walking up" the parents. */
    VNODE_T *pvp;               /* Local parent vp for "walking up". */
    tbs_boolean_t rmvd;         /* Input component removed or just refcnt'd. */
    int error;                  /* Error we will return. */

    /* Should only be called correctly. */
    ASSERT(dvp != NULL && nmp != NULL && vpp != NULL);

    /* Lookup the component (it should be there). */
    if ((error = mvfs_auvw_lookup_comp_idx(dkey, &vpidx)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_unmount_comp: Lookup of nm=%s in dvp=%p err=%d\n",
                 nmp, dvp, error);
        goto done;
    }
    /* Hold the vp we found.  We might remove the entry below (which would rele
    ** the vp) and we need the vp to find the ".." entry for this component.
    ** Also (if not error) the caller expects it to be held.
    */
    vp = mvtdp->vobtag_dir_ents[vpidx].vobtag_vp;
    VN_HOLD(vp);

    /* We'll need the parent vp for our refcnt loop below, so remember it here
    ** before we (possibly) remove the entry. The pvp is held by its own entry
    ** so we don't need an extra hold on it (even though the remove may rele
    ** it).
    */
    pvp = mvtdp->vobtag_dir_ents[vpidx].vobtag_parent_vp;

    /* If we're not at the last component of the name, we just want the caller
    ** to keep parsing (we handle "backing up the path" for refcnts below if we
    ** actually remove the entry).
    */
    if (!lc) {
        goto done;
    }
    /* Remove the entry, and the associated ".." entry (if we actually remove
    ** the entry...see mvfs_auvw_mount_comp for why we have a separate ".."
    ** entry).
    */
    if ((error = mvfs_auvw_remove_comp(vp, vpidx, cdp, &rmvd)) != 0) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_unmount_comp: remove nm=%s (vp=%p) err=%d\n",
                 nmp, vp, error);
        goto done;
    }
    if (!rmvd) {
        /* We just refcnt'd the entry, so no parent updates needed. */
        goto done;
    }
    /* Continue "up the chain", removing parents to get the refcnts right.  We
    ** start with the parent vp we saved above and continue until we don't
    ** remove an entry or our parent is the view (for which we don't have an
    ** entry).
    */
    while (rmvd && !MFS_ISAUVIEW(VTOM(pvp))) {
        lvp = pvp;              /* Entry we're going to remove. */

        if ((error = mvfs_auvw_lookup_vpidx(lvp, &vpidx)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_unmount_comp: vpidx for lvp=%p pvp=%p err=%d\n",
                     lvp, pvp, error);
            goto done;
        }
        /* Remember our parent before removing our entry. */
        pvp = mvtdp->vobtag_dir_ents[vpidx].vobtag_parent_vp;

        if ((error = mvfs_auvw_remove_comp(lvp, vpidx, cdp, &rmvd)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_unmount_comp: remove nm=%s (vp=%p) err=%d\n",
                     mvtdp->vobtag_dir_ents[vpidx].vobtag_nm, lvp, error);
            goto done;
        }
    }
  done:
    /* On error, we need to rele the vp if we looked it up. */
    if (error != 0) {
        if (vp != NULL) {
            ATRIA_VN_RELE(vp, cdp);
            vp = NULL;
        }
    }
    /* The vp was held above, or it is NULL. */
    *vpp = vp;
    return error;
}

/****************************************************************************
 * mvfs_auvw_parse_vobtag_init
 * Check and initialize tagp to point to the first component in a vobtag.
 * IO   tagpp           Pointer to start of vobtag to parse
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_parse_vobtag_init(char **tagpp)
{
    char *tagp = *tagpp;        /* Local var moves. */
    int error;

    /* Make sure there's something here to parse. */
    if ((tagp == NULL) || (*tagp == '\0')) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_parse_vobtag_init: tagp=%p is NULL"
                 " or an empty string and we expect a vobtag.\n",
                 *tagpp);
        error = EINVAL;
    } else {
        /* Skip past initial path separator(s) (our parsing loop expects no
        ** starting separator).
        */
        while (PN_IS_SEPCHAR(*tagp)) {
            tagp++;
        }
        /* If there are only separators, it's an error. */
        if (*tagp == '\0') {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_parse_vobtag_init: The vobtag is only"
                     " separators: %s\n",
                     *tagpp);
            error = EINVAL;
        } else {
            /* Looks good, caller should start from here. */
            *tagpp = tagp;
            error = 0;
        }
    }
    return error;
}

/****************************************************************************
 * mvfs_auvw_parse_vobtag
 * Return the next component into cnmp starting at tagp and advance tagp.
 * IO   tagpp           Pointer to start of vobtag to parse
 * IN   cnmp            Pointer to a buffer to hold the component found
 * IN   cnmsz           Size of the buffer pointed to by cnmp
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_parse_vobtag(
    char **tagpp,
    char *cnmp,
    size_t cnmsz
)
{
    char *tagp = *tagpp;        /* Local var moves. */
    size_t ccnt;                /* Character count in component. */
    int error;

    /* tagp points to the start of a component name.  Copy chars until the next
    ** separator or the end of the string (use array indexing since we need cnmp
    ** to point to the start of the name).  And, leave room at the end for a
    ** final '\0' char.

    ** XXXODFS what should we do about multi-byte characters, which are handled
    ** by the (libatriaks) strutl function in other code, but which we can't use
    ** here?  Since other code, e.g., mvfs_findvfs_mntpath, doesn't seem to
    ** worry about it, it's probably OK here, too.
    */
    for (ccnt = 0; ccnt < (cnmsz - 1); ccnt++) {
        if (PN_IS_SEPCHAR(*tagp) || *tagp == '\0') {
            break;              /* We're done with this component. */
        }
        cnmp[ccnt] = *tagp++;   /* Copy a char to the buffer. */
    }
    cnmp[ccnt] = '\0';      /* Always correctly terminate the name. */

    /* Check how we exited the loop (too long, or normally). */
    if (ccnt >= (cnmsz - 1)) {
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_parse_vobtag: component name: %s"
                 " in vobtag too long (more than %d chars).\n",
                 cnmp, cnmsz);
        error = ENAMETOOLONG;
    } else {
        /* Advance tagp past any separators at the end of the component we just
        ** copied to be ready for the next component, if there is one.
        */
        while (PN_IS_SEPCHAR(*tagp)) {
            tagp++;
        }
        /* Tell the caller where the next component starts. */
        *tagpp = tagp;
        error = 0;
    }
    return error;
}

/****************************************************************************
 * mvfs_auvw_mnt_umnt_vobtag
 * Create or remove "directories" for the components of a vobtag.
 * IN   vfsp            VFS for the VOB root (from mvfs_vmount_subr)
 * IN   cdp             Call data (includes user cred)
 * IN   mnt             TRUE - mount, FALSE - unmount
 * RESULT:              Unix errno
 */
STATIC int MVFS_NOINLINE
mvfs_auvw_mnt_umnt_vobtag(
    VFS_T *vfsp,
    CALL_DATA_T *cdp,
    tbs_boolean_t mnt
)
{
    size_t nmsz = sizeof(ks_name_t);
    int cmpcnt;                 /* Number of components we've seen. */
    tbs_boolean_t lc;           /* Last component in vobtag. */
    mfs_pn_char_t *tagp;        /* Current position in the vobtag. */
    VNODE_T *vp = NULL;         /* Vnode for a component. */
    VNODE_T *vwvp = NULL;       /* Vnode of view. */
    mvfs_vobtag_dir_ent_key_t dkey = {NULL, NULL}; /* Key of component. */
    mvfs_vobtag_dir_ent_key_t pdkey = {NULL, NULL}; /* Key of parent in dkey. */
    char *mnt_type = (mnt ? "(mount)" : "(unmount)");
    int error;

    /* Sanity check our inputs. */
    ASSERT((vfsp != NULL) && (VFS_TO_MMI(vfsp)->mmi_viewvp != NULL));
    tagp = VFS_TO_MMI(vfsp)->mmi_mntpath;

    if ((error = mvfs_auvw_parse_vobtag_init(&tagp)) != 0) {
        goto done;
    }
    /* Allocate space to hold each component as we find it.  If the name is used
    ** in an entry, it will be copied, so we can reuse it for the next
    ** component.  We also need to save the name as the name of the previous
    ** component, so we need another space allocation.
    */
    if ((dkey.vtk_nm = (char *)KMEM_ALLOC(nmsz, KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto done;
    }
    if ((pdkey.vtk_nm = (char *)KMEM_ALLOC(nmsz, KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto done;
    }
    /* The initial parent dir vp is the view associated with this VFS, which is
    ** set in mvfs_vmount.  Note, it is held by our caller, but we hold it again
    ** below to make our logic clearer.
    */
    vwvp = VFS_TO_MMI(vfsp)->mmi_viewvp;
    ASSERT(MFS_VPISMFS(vwvp) && MFS_ISAUVIEW(VTOM(vwvp)));

    /* Get the database lock for this view while we add/remove all the
    ** components in the name.  This makes it possible to manage the refcnts and
    ** cleanup on errors without worrying about things changing in the database.
    */
    MVFS_AUVW_VOBTAG_LOCK(MVFS_AUVW_VOBTAG_DIRP(VTOM(vwvp)));

    /* Parse the vobtag and process each component (init guaranteed there is at
    ** least one).  Hold the initial dir vp so the loop can manage all the holds
    ** and reles for the components.  The initial dir name could be the view,
    ** but we don't have any entry for the view, so we don't need it for lookup,
    ** so just make it empty.  We limit the component count to a maximum in case
    ** something goes wrong in the parsing.
    */
    vp = vwvp;
    VN_HOLD(vp);
    pdkey.vtk_parent_vp = NULL; /* No parent for the view vp. */
    pdkey.vtk_nm[0] = '\0';
    lc = FALSE;                 /* Not the last component yet. */

    for (cmpcnt = 1; cmpcnt <= MVFS_MAX_VOBTAG_COMP && !lc; cmpcnt++) {
        dkey.vtk_nm[0] = '\0';   /* Component starts empty. */
        dkey.vtk_parent_vp = vp; /* Dir vp in which to lookup component. */
        vp = NULL;               /* In case of error. */

        if ((error = mvfs_auvw_parse_vobtag(&tagp, dkey.vtk_nm, nmsz)) != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_mnt_umnt_vobtag: %s Parsing vobtag=%s\n",
                     mnt_type, VFS_TO_MMI(vfsp)->mmi_mntpath);
            goto done;
        }
        /* Set the last component flag (if we're at the end of the string) so
        ** mvfs_auvw_mount_comp knows if the component is a MFS_VOBTAGCLAS name
        ** or a MFS_VOBRTCLAS name.  It also ends our parsing loop.
        */
        lc = (*tagp == '\0');

        /* Create or remove the entry for this component.  The dkey provides the
        ** parent dir and name, and the output vp is created for that key and is
        ** returned held.  The pdkey provides the parent dir's parent dir and
        ** the parent dir's name so the parent dir can be found for refcnt
        ** purposes.
        */
        if (mnt) {
            error = mvfs_auvw_mount_comp(vfsp, pdkey, dkey, lc, cdp, &vp);
        } else {
            error = mvfs_auvw_unmount_comp(dkey, lc, cdp, &vp);
        }
        if (error != 0) {
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_mnt_umnt_vobtag: %s comp nm=%s in dvp=%p"
                     " (nm=%s) for vobtag=%s err=%d\n",
                     mnt_type, dkey.vtk_nm, dkey.vtk_parent_vp, pdkey.vtk_nm,
                     VFS_TO_MMI(vfsp)->mmi_mntpath, error);
            goto done;
        }
        /* The parent key becomes the current key, and the current key gets set
        ** at the top of the loop (to the vp we got back and the next component
        ** name we parse).
        */
        if (pdkey.vtk_parent_vp != NULL) {
            ATRIA_VN_RELE(pdkey.vtk_parent_vp, cdp);
        }
        pdkey.vtk_parent_vp = dkey.vtk_parent_vp;
        dkey.vtk_parent_vp = NULL; /* We transferred "ownership" to pdkey. */
        STRCPY(pdkey.vtk_nm, dkey.vtk_nm);
    }
    /* On loop exit (too many components or last component), dkey.vtk_parent_vp
    ** is NULL, and we're done with vp (which must be non-NULL because there was
    ** no error) and the parent's parent vp.
    */
    ASSERT((dkey.vtk_parent_vp == NULL) && (vp != NULL));

    if (pdkey.vtk_parent_vp != NULL) {
        ATRIA_VN_RELE(pdkey.vtk_parent_vp, cdp);
        pdkey.vtk_parent_vp = NULL;
    }
    ATRIA_VN_RELE(vp, cdp);
    vp = NULL;

    if (!lc) {
        /* Oops, exit from loop because of too many components. */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_mnt_umnt_vobtag: %s More than %d components"
                 " in vobtag: %s\n",
                 mnt_type,
                 MVFS_MAX_VOBTAG_COMP,
                 VFS_TO_MMI(vfsp)->mmi_mntpath);
        error = ENAMETOOLONG;
        goto done;
    }
  done:
    if (error != 0) {
        /* We could leave the loop with parent vp's held on error (but not vp,
        ** which should be NULL).  Trying to cleanup the names we (might) have
        ** put in already is just too hard, so just leave them there and hope it
        ** doesn't happen too often ;)
        */
        ASSERT(vp == NULL);

        if (dkey.vtk_parent_vp != NULL) {
            ATRIA_VN_RELE(dkey.vtk_parent_vp, cdp);
            dkey.vtk_parent_vp = NULL;
        }
        if (pdkey.vtk_parent_vp != NULL) {
            ATRIA_VN_RELE(pdkey.vtk_parent_vp, cdp);
            pdkey.vtk_parent_vp = NULL;
        }
    }
    /* We can get here on an error before we get the lock, so check first. */
    if (vwvp != NULL) {
        MVFS_AUVW_VOBTAG_UNLOCK(MVFS_AUVW_VOBTAG_DIRP(VTOM(vwvp)));
    }
    if (dkey.vtk_nm != NULL) {
        KMEM_FREE(dkey.vtk_nm, nmsz);
    }
    if (pdkey.vtk_nm != NULL) {
        KMEM_FREE(pdkey.vtk_nm, nmsz);
    }
    return error;
}

/****************************************************************************
 * mvfs_auvw_mount_vobtag
 * Create "directories" for the components of a vobtag.
 * IN   vfsp            VFS for the VOB root (from mvfs_vmount_subr)
 * IN   cdp             Call data (includes user cred)
 * RESULT:              Unix errno
 */
int
mvfs_auvw_mount_vobtag(
    VFS_T *vfsp,
    CALL_DATA_T *cdp
)
{
    return mvfs_auvw_mnt_umnt_vobtag(vfsp, cdp, TRUE);
}

/****************************************************************************
 * mvfs_auvw_unmount_vobtag
 * Remove "directories" for the components of a vobtag.
 * IN   vfsp            VFS for the VOB root (from mvfs_unregister_mount)
 * IN   cdp             Call data (includes user cred)
 * RESULT:              Unix errno
 */
int
mvfs_auvw_unmount_vobtag(
    VFS_T *vfsp,
    CALL_DATA_T *cdp
)
{
    return mvfs_auvw_mnt_umnt_vobtag(vfsp, cdp, FALSE);
}

/****************************************************************************
 * mvfs_auvw_getattr
 * Return canned vnode attributes structure for a VOBTAG object.  A NTVW or
 * VOBRT object will continue to get attrs the way it currently does (even
 * though we have a vob root entry in our database, it is only for looking up
 * ".." in an automatic view).
 * IN   vp              Vnode pointer to a view or vobtag object
 * IN   vap             (OPTIONAL) Ptr to a vnode attr structure to fill in
 * IN   cred            Ptr to the current user-credentials structure
 * RESULT:              Unix errno
 */
int
mvfs_auvw_getattr(
    VNODE_T *vp,
    VATTR_T *vap,
    CRED_T *cred
)
{
    mfs_mnode_t *mnp = VTOM(vp);
    mfs_mnode_t *vwmnp;
    dev_t dev;
    int error;

    ASSERT(MFS_ISVOBTAG(mnp));

    /* We use info from the containing view (since we don't really have any
    ** independent times, sizes, etc.).
    */
    vwmnp = VTOM(MFS_VIEW(vp));

    /* This routine is easy if no info is wanted. */
    if (vap != NULL) {
        /* See mvfs_ntvw_getattr() in mvfs_ntvwops.c for more info on these
        ** "canned" values.
        */
        VATTR_SET_TYPE(vap, VDIR);
        VATTR_SET_MODE_RIGHTS(vap, (S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH));
        VATTR_SET_MODE_TYPE(vap, S_IFDIR);
        MVFS_COPY_UID_TO_VATTR(vap, &(vwmnp->mn_view.cuid), vwmnp, &error);
        MVFS_COPY_GID_TO_VATTR(vap, &(vwmnp->mn_view.cgid), vwmnp, &error);
        dev = FSID_TO_DEV(VFS_FSID(vp->v_vfsp));
        VATTR_SET_FSID(vap, &dev);
        VATTR_SET_NODEID(vap, mnp->mn_hdr.mnum);
        VATTR_SET_NLINK(vap, 2);
        VATTR_SET_SIZE(vap, DEV_BSIZE);
        VATTR_SET_BLKSIZE(vap, DEV_BSIZE);
        VATTR_SET_ATIME_TS(vap, &(vwmnp->mn_view.ctime));
        VATTR_SET_MTIME_TS(vap, &(vwmnp->mn_view.ctime));
        VATTR_SET_CTIME_TS(vap, &(vwmnp->mn_view.ctime));
        VATTR_SET_RDEV(vap, (dev_t)(dev & 0xffff));
        VATTR_SET_NBLOCKS(vap, VATTR_BTODB(VATTR_GET_SIZE(vap)));
        VATTR_FILL(vap);
    }
    return 0;
}

/****************************************************************************
 * mvfs_auvw_lookup
 * Lookup nm in input dir vp and return a VOBRT, VOBTAG, or NTVIEW vp.
 * IN   dvp             Dir vnode pointer (NT viewtag, vobtag, or vobrt)
 * IN   nm              Name of the vobtag component to lookup
 * OUT  vpp             Resulting vnode ptr or NULL if the lookup fails
 * IO   pnp             Ptr to a pathname struct (may be NULL)
 * IN   cd              Ptr to call data (includes current user creds)
 * RESULT:              Unix errno
 */
int
mvfs_auvw_lookup(
  VNODE_T *dvp,
  char *nm,
  VNODE_T **vpp,
  struct pathname *pnp,
  CALL_DATA_T *cd
)
{
    mfs_mnode_t *mnp = VTOM(dvp);
    VNODE_T *vp = NULL;         /* Compiler can't see it's always set. */
    mvfs_vobtag_dir_ent_key_t dkey;  /* Key to lookup component. */
    int error;

    ASSERT(MVFS_VIEW_IS_AUTO_VIEW(dvp));

    /* Make sure there's something to lookup (STRCMP is unhappy with NULL). */
    if (nm == NULL) {
        error = EINVAL;
        goto done;
    }
    /* This routine handles only these mnode classes as "directories" and can
    ** only lookup ".." in a MFS_VOBRTCLAS.
    */
    if (!(MFS_ISAUVIEW(mnp) ||
          MFS_ISVOBTAG(mnp) ||
          (MFS_ISVOBRT(mnp) && (STRCMP(nm, "..") == 0))))
    {
        /* No lookup possible by this routine. */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_lookup: called with mclass=%d nm=%s"
                 " (only %d, %d, or %d && nm=.. are allowed).\n",
                 mnp->mn_hdr.mclass, nm,
                 MFS_NTVWCLAS, MFS_VOBTAGCLAS, MFS_VOBRTCLAS);
        vp = NULL;
        error = ENOENT;
        goto done;
    }
    /* For any dir type, "." is easy, so don't bother looking it up below. */
    if (STRCMP(nm, ".") == 0) {
        vp = dvp;
        VN_HOLD(vp);
        error = 0;
        goto done;
    }
    /* If the dir is a MFS_NTVWCLAS then there are some special lookups we need
    ** to do since we don't have an entry for the view in our database.  For
    ** MFS_VOBTAGCLAS and MFS_VOBRTCLAS, ".." was setup at VOB mount time for
    ** each vobtag component (treated as a dir), so it is just another lookup in
    ** our database below.
    */
    if (MFS_ISNTVIEW(mnp)) {
        if ((STRCMP(nm, "..") == 0)) {
            if ((vp = MVFS_GETVIEWROOT(MVFS_VRTYPE(mnp->mn_view.auto_view)))
                != NULL)
            {
                error = 0;          /* Getviewroot already did a hold. */
            } else {
                error = ENOENT;     /* The viewroot is not set up. */
            }
            goto done;
        }
        /* Handle SETVIEW^ hack...see mvfs_lookup_ctx.
        **
        ** Note, mvfs_ntvwops.c does this, too, but I'm not sure why since
        ** Windows doesn't have a setview.  Maybe because we allow Unix to use
        ** MFS_NTVWCLAS (see mvfs_viewroot_common_init in mvfs_vfsops.c),
        ** although in that case I don't think we allow setview there, either.
        ** Another question...should we handle other ^ names here?
        */
        if (PN_STRCMP(MVFS_PN_CI_LOOKUP(pnp), nm, "SETVIEW^") == 0) {
            VN_HOLD(dvp);       /* The next setview call uses up a hold. */

            if ((error = MVFS_SET_PROCVIEW(dvp, NULL, cd)) == 0) {
                VN_HOLD(dvp);   /* We're returning this, so hold it. */
                vp = dvp;
            }
            goto done;
        }
    }
    /* Lookup the name in the directory we got.  vp is always set and is
    ** returned held if it is not NULL.  If it is NULL then error != 0.  We need
    ** to use MFS_VIEW(dvp) to get the right lock because the dvp we have might
    ** be a view itself or it might be a vobtag or vobrt object, in which case
    ** we want to use the view it is associated with (rather than the object
    ** itself).
    */
    dkey.vtk_parent_vp = dvp;
    dkey.vtk_nm = nm;

    MVFS_AUVW_VOBTAG_LOCK(MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp))));

    error = mvfs_auvw_lookup_comp(dkey, pnp, &vp);

    MVFS_AUVW_VOBTAG_UNLOCK(MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp))));

done:
    *vpp = vp;               /* vp has been set and held appropriately above. */
    return error;
}

/****************************************************************************
 * mvfs_auvw_dirent
 * This routine initializes a dirent and decides if it is OK and adds it to the
 * user's buffer if so.  It can return with error 0 and uio_fullp TRUE (or
 * FALSE) and *offsetp will have been updated appropriately.  If error != 0,
 * then *offsetp won't be changed.
 * IN   nm              The name in the dirent.
 * IN   direntp         Pointer to a dirent to fill in (max size pathname).
 * IN   inode           The inode in the dirent.
 * IN   uiop            Pointer to the UIO structure for the user's buffer.
 * IN   orig_uio_resid  Value of uiop->uio_resid before this call.
 * IO   offsetp         Pointer to an offset that will be updated in some cases.
 * OUT  uio_fullp       TRUE - UIO buffer full.
 * RESULT:              Unix errno
 */
STATIC int
mvfs_auvw_dirent(
    char *nm,
    KDIRENT_T *direntp,
    u_long inode,
    struct uio *uiop,
    MVFS_UIO_RESID_T orig_uio_resid,
    tbs_boolean_t ext,
    MOFFSET_T *offsetp,
    tbs_boolean_t *uio_fullp
)
{
    size_t nmlen;
    size_t reclen;
    int error = 0;

    /* We don't check explicitly to see if offset has grown larger that
    ** MVFS_MAX_DIRSIZE because we limit the growth.
    */
    ASSERT(*offsetp <= MVFS_MAX_DIRSIZE);

    /* Start not full, and set TRUE in some cases below. */
    *uio_fullp = FALSE;

    /* Skip any entry for which the name is too long (or NULL, which would panic
    ** the STRLEN).  This is a bug from the mount, and we don't want to even see
    ** this in the virtual dir, so we also don't update the offset.
    */
    if ((nm == NULL) || ((nmlen = STRLEN(nm)) > MAXNAMELEN)) {
        goto done;
    }
    /* Setup the dirent we (might) want to return. */
    if (ext) {
        MDKI_PANIC("mvfs_auvw_dirent: Called as mvfs_auvw_readdirx.");
    } else {
        reclen = KDIRENT_RECLEN(nmlen);
        KDIRENT_INIT((KDIRENT_T *)direntp,
                     inode, nm, nmlen, *offsetp + reclen);
    }
    /* If the offset in the virtual dir is before the user request, skip this
    ** entry, but update the offset for the next entry.
    */
    if (*offsetp < MVFS_UIO_OFFSET(uiop)) {
        *offsetp += reclen;
        goto done;
    }
    /* Check that the entry will fit in the uio buffer. */
    if (reclen > uiop->uio_resid) {
        /* Check if this is the first entry we want to return. */
        if (uiop->uio_resid == orig_uio_resid) {
            /* The first one doesn't fit, give up. */
            error = EINVAL;
        } else {
            /* We've already returned some so this is OK. */
            *uio_fullp = TRUE;
            error = 0;
        }
        goto done;
    }
    /* This call updates uiop->uio_resid (among other things). */
    if ((error = READDIR_UIOMOVE((caddr_t)direntp,
                                 &reclen,
                                 UIO_READ,
                                 uiop,
                                 *offsetp)) != 0)
    {
        goto done;
    }
    if (READDIR_BUF_FULL(uiop)) {
        *uio_fullp = TRUE;
    } else {
        /* Everything is OK, so update the offset for our caller. */
        *offsetp += reclen;
    }
  done:
    return error;
}

/****************************************************************************
 * mvfs_auvw_readdir_subr
 * Fill the user's buffer with dirents for the input directory.
 * IN   dvp             Vnode pointer to auview, vobtag, or vobrt component
 * IN   uiop            Ptr to uio struct decribing the user's buffer
 * IN   cd              Ptr to call data (includes current user-cred)
 * IN   ext             TRUE - readdirx, FALSE - readdir
 * RESULT:              Unix errno
 */
STATIC int
mvfs_auvw_readdir_subr(
  VNODE_T *dvp,
  struct uio *uiop,
  CALL_DATA_T *cd,
  tbs_boolean_t ext
)
{
    mfs_mnode_t *mnp;
    size_t direntlen;
    size_t dirent0len;
    void *direntp;
    MOFFSET_T offset;
    MVFS_UIO_RESID_T oresid;
    int error;
    int i;
    mvfs_vobtag_dir_t *mvtdp;
    tbs_boolean_t uio_full;
    int low;
    int high;
    int mid;
    int cmp;

    ASSERT(MFS_VPISMFS(dvp));
    mnp =  VTOM(dvp);
    ASSERT(MFS_ISNTVIEW(mnp) || MFS_ISVOBTAG(mnp) || MFS_ISVOBRT(mnp));

    if (ext) {
        MDKI_PANIC("mvfs_auvw_readdir_subr: Called as mvfs_auvw_readdirx.");
    } else {
        direntlen = KDIRENT_RECLEN(MAXNAMELEN);
        dirent0len = KDIRENT_RECLEN(0);
    }
    /* Initial checks */
    if (MVFS_IS_INVALID_OFFSET(MVFS_UIO_OFFSET(uiop)) ||
        (uiop->uio_resid <= dirent0len))
    {
        return ENOENT;
    }
    /* Create an internal dirent buffer to use that can hold a maximum length
    ** "name" so mvfs_auvw_dirent below doesn't have to allocate one each time.
    */
    if ((direntp = (void *)KMEM_ALLOC(direntlen, KM_SLEEP)) == NULL) {
        return ENOMEM;
    }
    /* Return as many items as the buffer will hold.  Note that at EOF we must
    ** return no entries with an error = 0.  See mvfs_ntvw_readdir in
    ** mvfs_ntvwops.c for more comments about various aspects of this code
    ** (byte-offsets vs. cookies, scanning from the beginning, etc.).
    */
    error = 0;
    oresid = uiop->uio_resid;   /* Starting resid before moving any data. */
    offset = 0;                 /* Offset in the virtual dir to start. */

    /* We generate a "." entry for everybody since it is easy and saves an entry
    ** in our database (for each dir there).  If the dvp is a MFS_NTVWCLAS then
    ** we don't have any ".." entry for it, either, since it never came through
    ** our mvfs_auvw_mount_vobtag routine.  For other dvp's we have a ".." entry
    ** in the database, so it will just get picked up in the normal traversal.
    ** Note, we use the mf_mnum version of the mfs_fid because the mf_dbid
    ** version is for VOB objects and these are not.
    */
    if (((error =
          mvfs_auvw_dirent(
              ".",
              direntp,
              VTOM(dvp)->mn_hdr.fid.mf_mnum,
              uiop,
              oresid,
              ext,
              &offset,
              &uio_full)) != 0)
        || uio_full)
    {
        goto done;
    }
    if (MFS_ISAUVIEW(mnp)) {
        /* Use the mnum for the automatic viewroot (since this is a view). */
        if (((error =
              mvfs_auvw_dirent(
                  "..",
                  direntp,
                  VTOM(MVFS_GETVIEWROOT(AUVWROOT))->mn_hdr.fid.mf_mnum,
                  uiop,
                  oresid,
                  ext,
                  &offset,
                  &uio_full)) != 0)
            || uio_full)
        {
            goto done;
        }
    }
    /* The primary key for out database is the directory vp and we're going to
    ** return all the entries with that primary key (which doesn't have to be
    ** unique).  Therefore, use a variation of a binary search that finds the
    ** leftmost entry for duplicated keys.  Note, in lookup, the combination of
    ** the primary and secondary keys (dvp and name) is unique, so this issue
    ** doesn't arise there.
    **
    ** Do a sequential scan from the leftmost entry we find to the rightmost
    ** entry with the same key.  The mvfs_auvw_dirent call below will start
    ** returning entries when we have passed the desired user offset and stop if
    ** the uio buffer fills up (the caller will call in for the next batch of
    ** entries with the offset of the next entry they need).  The ".." entry is
    ** in the database explicitly, so no special case is needed.
    */
    mvtdp = MVFS_AUVW_VOBTAG_DIRP(VTOM(MFS_VIEW(dvp)));
    MVFS_AUVW_VOBTAG_LOCK(mvtdp);

    if (mvtdp->vobtag_dir_count < 1) {
        /* Our database is empty, don't go looking for entries. */
        if (MFS_ISAUVIEW(mnp)) {
            /* For a view we've returned "." and ".." above and there is
            ** no more, which is OK.
            */
            error = 0;
        } else {
            /* It's not a view, there should be something in our database (at
            ** least a "..").
            */
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_auvw_readdir_subr: dvp=%p empty database.\n",
                     dvp);
            error = ENOENT;
        }
        goto unlock;
    }
    /* Find the leftmost index for our (non-unique) primary key.  We don't need
    ** an (int) cast for setting high here because we've guaranteed that the
    ** count is at least 1.
    */
    low = 0;
    high = mvtdp->vobtag_dir_count - 1;

    while (low < high) {
        mid = low + ((high - low) / 2);

        if (mvtdp->vobtag_dir_ents[mid].vobtag_parent_vp < dvp) {
            low = mid + 1;
        } else {
            high = mid;
        }
    }
    /* Verify we found the leftmost index. */
    if (!((low == high) &&
          (mvtdp->vobtag_dir_ents[low].vobtag_parent_vp == dvp)))
    {
        /* That's not good, and we already returned the "." entry. */
        mvfs_log(MFS_LOG_ERR,
                 "mvfs_auvw_readdir_subr: dvp=%p not in database.\n",
                 dvp);
        error = ENOENT;
        goto unlock;
    }
    /* Just scan from left to right for matching dvp entries. */
    for (i = low; mvtdp->vobtag_dir_ents[i].vobtag_parent_vp == dvp; i++) {

        /* We have the per-view database locked, but there's no reason the
        ** UIOMOVE in the following call should run into this lock (as opposed
        ** to mvfs_ntvw_readdir where the lock is on a global table) so don't
        ** unlock/relock here.
        */
        if (((error =
              mvfs_auvw_dirent(
                  mvtdp->vobtag_dir_ents[i].vobtag_nm,
                  direntp,
                  VTOM(mvtdp->vobtag_dir_ents[i].vobtag_vp)->mn_hdr.fid.mf_mnum,
                  uiop,
                  oresid,
                  ext,
                  &offset,
                  &uio_full)) != 0)
            || uio_full)
        {
            break;
        }
    }
  unlock:
    MVFS_AUVW_VOBTAG_UNLOCK(mvtdp);

  done:
    KMEM_FREE(direntp, direntlen);
    return error;
}

/****************************************************************************
 * mvfs_auvw_readdir
 * Fill the user's buffer with dirents for the input directory.
 * IN   dvp             Vnode pointer to auview, vobtag, or vobrt component
 * IN   uiop            Ptr to uio struct decribing the user's buffer
 * IN   cd              Ptr to call data (includes current user-cred)
 * RESULT:              Unix errno
 */
int
mvfs_auvw_readdir(
  VNODE_T *dvp,
  struct uio *uiop,
  CALL_DATA_T *cd
)
{
    return mvfs_auvw_readdir_subr(dvp, uiop, cd, FALSE);
}

/****************************************************************************
 * mvfs_vobtag_dir_init
 * Initializes the vobtag dir structure for the given mnode.
 * The mount_lock within the structure is initialized using the viewname for the
 * mnode.  The name of the lock is of the form "mount_lock [tag]".  The mnode is
 * assumed to be locked.
 * IN   mnp             The auto view mnode containing the vobtag dir
 * RESULT:              Unix error
 */
int
mvfs_vobtag_dir_init(mfs_mnode_t *mnp)
{
    char *mnt_lock_tag = mnp->mn_view.viewname;
    char *mnt_lock_name = NULL;
    size_t mnt_lock_name_size = 0;
    mvfs_vobtag_dir_t *mvtdp = NULL;
    mvfs_vobtag_dir_ent_t *mvtdep = NULL;
    int error;

    ASSERT(MISLOCKED(mnp) &&
           MVFS_VIEW_IS_AUTO_VIEW(MTOV(mnp)) &&
           (MVFS_AUVW_VOBTAG_DIRP(mnp) == NULL)
    );
    /* Allocate the vobtag directory structure for this view. */
    if ((mvtdp = KMEM_ALLOC(sizeof(*mvtdp), KM_SLEEP)) == NULL) {
        error = ENOMEM;
        goto done;
    }
    /* Allocate the vobtag dir entries separately. */
    if ((mvtdep = KMEM_ALLOC(sizeof(*mvtdep) * MVFS_VOBTAG_DIR_ENTS, KM_SLEEP))
        == NULL)
    {
        error = ENOMEM;
        goto done;
    }
    /* Finish initializing now that everything is allocated. */
    error = 0;
    mvtdp->vobtag_dir_count = 0;
    BZERO(mvtdep, sizeof(*(mvtdep)) * MVFS_VOBTAG_DIR_ENTS);

    /* Initialize the vobtag_dir_lock with a "nice" name. */
    mnt_lock_name_size =
        STRLEN("vobtag_lock [") +
        STRLEN(mnt_lock_tag) +
        2; /* Extra space for "]\0" */

    if ((mnt_lock_name = KMEM_ALLOC(mnt_lock_name_size, KM_SLEEP)) == NULL) {
        mnt_lock_name = "auvw_mount_lock"; /* Just pick a name. */
    }
    MVFS_SNPRINTF(mnt_lock_name,
                  (int)mnt_lock_name_size,
                  "%s [%s]",
                  "vobtag_lock",
                  mnt_lock_tag);
    mnt_lock_name[mnt_lock_name_size - 1] = '\0';
    INITLOCK(&(mvtdp->vobtag_dir_lock), mnt_lock_name);

    /* INITLOCK makes a copy of mnt_lock_name, so free it here since we no
    ** longer need it.
    */
    KMEM_FREE(mnt_lock_name, mnt_lock_name_size);

done:
    if (error != 0) {
        /* Cleanup whatever we allocated. */
        if (mvtdep != NULL) {
            KMEM_FREE(mvtdep, sizeof(*mvtdep) * MVFS_VOBTAG_DIR_ENTS);
        }
        if (mvtdp != NULL) {
            KMEM_FREE(mvtdp, sizeof(*mvtdp));
        }
    } else {
        /* Fill in the mnode with the vobtag directory information. */
        mvtdp->vobtag_dir_ents = mvtdep;
        MVFS_AUVW_VOBTAG_DIRP(mnp) = mvtdp;
    }
    return error;
}

/****************************************************************************
 * mvfs_vobtag_dir_clean
 * Cleans the vobtag directory for the given mnode.
 * The mnode is assumed to be locked.  Sets MVFS_AUVW_VOBTAG_DIRP(mnp) to NULL
 * before returning.
 * IN   mnp             The mnode containing the vobtag dir (must be an auto view mnode).
 * IN   cdp             Call data pointer.
 * RESULT:              void
 */
void
mvfs_vobtag_dir_clean(
    mfs_mnode_t *mnp,
    CALL_DATA_T *cdp
)
{
    mvfs_vobtag_dir_t *mvtdp = MVFS_AUVW_VOBTAG_DIRP(mnp);
    mvfs_vobtag_dir_ent_t *mvtdep = mvtdp->vobtag_dir_ents;
    int i;

    ASSERT(MISLOCKED(mnp) &&
           MVFS_VIEW_IS_AUTO_VIEW(MTOV(mnp)) &&
           (mvtdp != NULL) &&
           (mvtdep != NULL)
    );
    /* Cleanup everything we allocated or held.  It really should all be gone
    ** (i.e., vobtag_dir_count should be 0), but be safe.
    */
    for (i = 0; i < mvtdp->vobtag_dir_count; i++) {
        if (i == 0) {
            /* Just log this once. */
            mvfs_log(MFS_LOG_ERR,
                     "mvfs_vobtag_dir_clean: vobtag dir=%p cnt=%d for mnp=%p"
                     " not empty...continuing.\n",
                     mvtdp, mvtdp->vobtag_dir_count, mnp);
        }
        if (mvtdep[i].vobtag_nm != NULL) {
            KMEM_FREE(mvtdep[i].vobtag_nm, STRLEN(mvtdep[i].vobtag_nm));
        }
        if (mvtdep[i].vobtag_parent_vp != NULL) {
            ATRIA_VN_RELE(mvtdep[i].vobtag_parent_vp, cdp);
        }
        if (mvtdep[i].vobtag_vp != NULL) {
            ATRIA_VN_RELE(mvtdep[i].vobtag_vp, cdp);
        }
    }
    /* Free the vobtag_dir_ent storage. */
    KMEM_FREE(mvtdep, sizeof(*mvtdep) * MVFS_VOBTAG_DIR_ENTS);
    mvtdp->vobtag_dir_ents = NULL; /* Just for cleanliness (we free next). */

    /* Free the vobtag_dir storage. */
    FREELOCK(&(mvtdp->vobtag_dir_lock));
    KMEM_FREE(mvtdp, sizeof(*mvtdp));

    /* Set the mnode pointer to NULL now it's all gone. */
    MVFS_AUVW_VOBTAG_DIRP(mnp) = NULL;
}
static const char vnode_verid_mvfs_auvwops_c[] = "$Id:  be0c6eb5.52a411e6.8dde.44:37:e6:71:22:3b $";
